/*
 * Simulator of MCS51 (dump.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "i_string.h"

#include "simcl.h"

#include "uc51cl.h" //FIXME
#include "globals.h"
#include "cmdutil.h"


/*
 * Dump address and instruction at specified addr
 */
//FIXME
void
dump_disass(class t_uc51 *uc, uint addr, class cl_sim *sim)
{
  char *dis;
  class cl_brk *b;
  int i;

  b  = uc->fbrk_at(addr);
  dis= uc->disass(addr, NULL);
  if (b)
    fprintf(sim->cmd_out, "%c", (b->perm == brkFIX)?'F':'D');
  else
    fprintf(sim->cmd_out, " ");
  fprintf(sim->cmd_out, "%c %06x %02x",
	 uc->inst_at(addr)?' ':'*',
	 addr, uc->get_mem(MEM_ROM, addr));
  for (i= 1; i < uc->inst_length(uc->get_mem(MEM_ROM, addr)); i++)
    fprintf(sim->cmd_out, " %02x", uc->get_mem(MEM_ROM, addr+i));
  while (i < 3)
    {
      fprintf(sim->cmd_out, "   ");
      i++;
    }
  fprintf(sim->cmd_out, " %s\n", dis);
  free(dis);
}


/*
 * Dump memory
 */

void
dump_memory(uchar *buf, int size, uint *start, uint stop, int bpl, FILE *f,
	    class cl_sim *sim)
{
  int i, idx;

  while ((*start <= stop) &&
	 (*start < size))
    {
      fprintf(f, "%06x ", *start);
      for (i= 0; (i < bpl) &&
	     (*start+i < size) &&
	     (*start+i <= stop);
	   i++)
	fprintf(f, "%02x ", buf[*start+i]);
      while (i < bpl)
	{
	  fprintf(f, "   ");
	  i++;
	}
      for (i= 0; (i < bpl) &&
	     (*start+i < size) &&
	     (*start+i <= stop);
	   i++)
	fprintf(f, "%c", isprint(buf[*start+i])?buf[*start+i]:'.');
      fprintf(f, "\n");
      (*start)+= bpl;
    }
}


/*
 * DISASSEMBLE [start [offset [lines]]]
 */
//FIXME
static disass_last_stop= 0;

bool
cmd_disass(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start, offset= -1, dir;
  int  lines= 20;
  uint realstart;

  if (!uc->there_is_inst())
    return(FALSE);
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= strtol(s, NULL, 0);
  else
    start= disass_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    offset= strtol(s, NULL, 0);
  if ((s= strtok(NULL, delimiters)) != NULL)
    lines= strtol(s, NULL, 0);

  realstart= start;
  while (!uc->inst_at(realstart))
    realstart= (realstart+1) & (EROM_SIZE-1);
  if (offset)
    {
      dir= (offset < 0)?-1:+1;
      while (offset)
	{
	  realstart= (realstart+dir) & (EROM_SIZE-1);
	  while (!uc->inst_at(realstart))
	    realstart= (realstart+dir) & (EROM_SIZE-1);
	  offset+= -dir;
	}
    }
  
  while (lines)
    {
      dump_disass(uc, realstart, sim);
      realstart= (realstart+1) & (EROM_SIZE-1);
      while (!uc->inst_at(realstart))
	realstart= (realstart+1) & (EROM_SIZE-1);
      lines--;
    }

  disass_last_stop= realstart;
  return(FALSE);
}


/*
 * DUMP CODE [start [stop]]
 */
//FIXME
static int dc_last_stop= 0;

bool
cmd_dump_code(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  int  lines= 20;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dc_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    {
      stop = (int)strtol(s, NULL, 0);
      lines= -1;
    }
  else
    stop = -1;
  while (((stop >= 0) && (start <= stop)) ||
	 (lines >= 0))
    {
      dump_disass(uc, start, sim);
      start= (start+uc->inst_length(uc->get_mem(MEM_ROM, start)))&(EROM_SIZE-1);
      lines--;
    }
  dc_last_stop= start;
  return(FALSE);
}


/*
 * DUMP CODE HEX [start [stop]]
 */

static int dch_last_stop= 0;

bool
cmd_dump_code_hex(char *cmd, class cl_uc *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dch_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop = (int)strtol(s, NULL, 0);
  else
    stop = start + (10*16) - 1;
  dump_memory(uc->MEM(MEM_ROM), EROM_SIZE, (uint *)&start, stop, 16,
	      sim->cmd_out, sim);
  dch_last_stop= start;
  return(FALSE);
}


/*
 * DUMP IRAM [start [stop]]
 */
//FIXME
static int di_last_stop= 0;

bool
cmd_dump_iram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= di_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop = (int)strtol(s, NULL, 0);
  else
    stop = start + (10*8) - 1;
  dump_memory(uc->IRAM, IRAM_SIZE, (uint *)&start, stop, 8,
	      /*uc*/sim->cmd_out, sim);
  di_last_stop= start;
  return(FALSE);
}


/*
 * DUMP XRAM [start [stop]]
 */
//FIXME
static int dx_last_stop= 0;

bool
cmd_dump_xram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dx_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop= (int)strtol(s, NULL, 0);
  else
    stop= start + (10*16) - 1;
  uc->eram2xram();
  dump_memory(uc->MEM(MEM_XRAM), uc->get_mem_size(MEM_XRAM),
	      (uint *)&start, stop, 16,
	      sim->cmd_out, sim);
  dx_last_stop= start;
  return(FALSE);
}


/*
 * DUMP REGISTERS
 */
//FIXME
bool
cmd_dump_regs(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  uint  start;
  uchar data;

  start= uc->SFR[PSW] & 0x18;
  dump_memory(uc->IRAM, IRAM_SIZE, &start, start+7, 8, /*uc*/sim->cmd_out, sim);
  start= uc->SFR[PSW] & 0x18;
  data= uc->IRAM[uc->IRAM[start]];
  fprintf(/*uc*/sim->cmd_out, "%06x %02x %c",
	  uc->IRAM[start], data, isprint(data)?data:'.');

  fprintf(/*uc*/sim->cmd_out, "  ACC= 0x%02x %3d %c  B= 0x%02x",
	  uc->SFR[ACC], uc->SFR[ACC], isprint(uc->SFR[ACC])?(uc->SFR[ACC]):'.',
	  uc->SFR[B]); 
  uc->eram2xram();
  data= uc->get_mem(MEM_XRAM, uc->SFR[DPH]*256 + uc->SFR[DPL]);
  fprintf(/*uc*/sim->cmd_out, "   DPTR= 0x%02x%02x @DPTR= 0x%02x %3d %c\n",
	  uc->SFR[DPH], uc->SFR[DPL],
	  data, data, isprint(data)?data:'.');

  data= uc->IRAM[uc->IRAM[start+1]];
  fprintf(/*uc*/sim->cmd_out, "%06x %02x %c", uc->IRAM[start+1], data,
	  isprint(data)?data:'.');
  data= uc->SFR[PSW];
  fprintf(/*uc*/sim->cmd_out, "  PSW= 0x%02x CY=%c AC=%c OV=%c P=%c\n", data,
	  (data&bmCY)?'1':'0', (data&bmAC)?'1':'0', (data&bmOV)?'1':'0',
	  (data&bmP)?'1':'0');

  dump_disass(uc, uc->PC, sim);
  return(FALSE);
}


/*
 * DUMP PORTS
 */
//FIXME
bool
cmd_dump_port(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  uchar data;

  data= uc->SFR[P0];
  fprintf(/*uc*/sim->cmd_out, "P0    ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data,
	  isprint(data)?data:'.');

  data= uc->SFR[P1];
  fprintf(/*uc*/sim->cmd_out, "    P1    ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data,
	  isprint(data)?data:'.');

  data= uc->port_pins[0];
  fprintf(/*uc*/sim->cmd_out, "Pin0  ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data,
	  isprint(data)?data:'.');

  data= uc->port_pins[1];
  fprintf(/*uc*/sim->cmd_out, "    Pin1  ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[0] & uc->SFR[P0];
  fprintf(/*uc*/sim->cmd_out, "Port0 ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[1] & uc->SFR[P1];
  fprintf(/*uc*/sim->cmd_out, "    Port1 ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  fprintf(/*uc*/sim->cmd_out, "\n");

  data= uc->SFR[P2];
  fprintf(/*uc*/sim->cmd_out, "P2    ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->SFR[P3];
  fprintf(/*uc*/sim->cmd_out, "    P3    ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[2];
  fprintf(/*uc*/sim->cmd_out, "Pin2  ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[3];
  fprintf(/*uc*/sim->cmd_out, "    Pin3  ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[2] & uc->SFR[P2];
  fprintf(/*uc*/sim->cmd_out, "Port2 ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[3] & uc->SFR[P3];
  fprintf(/*uc*/sim->cmd_out, "    Port3 ");
  print_bin(data, 8, /*uc*/sim->cmd_out);
  fprintf(/*uc*/sim->cmd_out, " 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  return(FALSE);
}


/*
 * DUMP SFR [addr...]
 */
//FIXME
bool
cmd_dump_sfr(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  uint start= 0x80;
  uchar data;
  struct name_entry *ne;

  if ((s= strtok(NULL, delimiters)) != NULL)
    while (s != NULL)
      {
	if ((ne= get_name_entry(sfr_tab, s, uc)) != NULL)
	  {
	    data= uc->SFR[ne->addr];
	    fprintf(/*uc*/sim->cmd_out, "%6s %02x %3d %c\n", ne->name, data, data,
		    isprint(data)?data:'.');
	  }
	else
	  {
	    start= strtol(s, NULL, 0);
	    data = uc->SFR[start];
	    if (start >= SFR_START)
	      fprintf(/*uc*/sim->cmd_out, "%06x %02x %3d %c\n", start, data, data,
		      isprint(data)?data:'.');
	  }
	s= strtok(NULL, delimiters);
      }
  else
    // dump all
    dump_memory(uc->SFR, SFR_SIZE, &start, 255, 16, /*uc*/sim->cmd_out, sim);
  return(FALSE);
}


/*
 * DUMP BIT addr...
 */
//FIXME
bool
cmd_dump_bit(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s, *sym;
  uchar *cell, addr;
  uchar bitaddr, bitmask;

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Address has not given.\n");
      return(FALSE);
    }
  while (s)
    {
      if (!interpret_bitname(s, uc, &cell, &addr, &bitaddr, &bitmask, &sym))
	{
	  fprintf(/*uc*/sim->cmd_out, "Bad address %s\n", s);
	  return(FALSE);
	}
      fprintf(/*uc*/sim->cmd_out, "%06x %6s %c\n",
	      addr, sym, 
	      (*cell & bitmask)?'1':'0');
      free(sym);
      s= strtok(NULL, delimiters);
    }
  return(FALSE);
}


/* End of dump.cc */
