/*
 * Simulator of MCS51 (set.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "i_string.h"

#include "stypes.h"
#include "simcl.h"

#include "uc51cl.h" //FIXME
#include "globals.h"
#include "cmdutil.h"
#include "dump.h"


/*
 * Parsing command and write data into specified buffer
 */
//FIXME
static void
set_memory(uchar *what, int size, int first,
	   class t_uc51 *uc, struct name_entry tabl[], class cl_sim *sim)
{
  uint start, addr;
  char *s, *p;
  uchar data;
  struct name_entry *ne= NULL;

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Address has not given.\n");
      return;
    }
  if (tabl)
    ne= get_name_entry(tabl, s, uc);
  if (ne)
    addr= ne->addr;
  else
    {
      addr= strtol(s, &p, 0);
      if (p &&
	  *p)
	{
	  fprintf(/*uc*/sim->cmd_out, "Bad address.\n");
	  return;
	}
    }
  start= addr;
  if (start >= size)
    {
      fprintf(/*uc*/sim->cmd_out,
	      "Address %d(0x%x) is bigger than %d(0x%x).\n",
	      start, start, size-1, size-1);
      return;
    }
  if (!(start >= first))
    {
      fprintf(/*uc*/sim->cmd_out,
	      "Address %d(0x%x) is less than %d(0x%x).\n",
	      start, start, first, first);
      return;
    }
  if ((s= strtok(NULL, " \t\v")) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Data has not given.\n");
      return;
    }
  while (s &&
	 addr < size)
    {
      if (isdigit(*s))
	{
	  data= strtol(s, &p, 0);
	  if (p &&
	      *p)
	    {
	      fprintf(/*uc*/sim->cmd_out, "Bad data %s\n", s);
	      break;
	    }
	  what[addr]= data;
	  if (what == uc->SFR)
	    {
	      uc->proc_write(&what[addr]);
	      if (addr == ACC)
		uc->set_p_flag();
	    }
	  if (what == uc->MEM(MEM_ROM))
	    uc->del_inst_at(addr);
	  addr++;
	}
      else
	{
	  int i, len;
	  p= proc_escape(s, &len);
	  i= 0;
	  while ((i < len) &&
		 addr < size)
	    {
	      what[addr]= p[i++];
	      if (what == uc->SFR)
		{
		  uc->proc_write(&what[addr]);
		  if (addr == ACC)
		    uc->set_p_flag();
		}
	      if (what == uc->MEM(MEM_ROM))
		uc->del_inst_at(addr);
	      addr++;
	    }
	  free(p);
	}
      s= strtok(NULL, " \t\v");
    }
  dump_memory(what, size, &start, addr-1, 16, /*uc*/sim->cmd_out, sim);
}


/*
 * SET IRAM addr data...
 */
//FIXME
bool
cmd_set_iram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  set_memory(uc->IRAM, IRAM_SIZE, 0, uc, NULL, sim);
  return(FALSE);
}


/*
 * SET XRAM addr data...
 */
//FIXME
bool
cmd_set_xram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  uc->eram2xram();
  set_memory(uc->MEM(MEM_XRAM), uc->get_mem_size(MEM_XRAM), 0, uc, NULL, sim);
  uc->xram2eram();
  return(FALSE);
}


/*
 * SET CODE addr data...
 */
//FIXME
bool
cmd_set_code(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  set_memory(uc->MEM(MEM_ROM), EROM_SIZE, 0, uc, NULL, sim);
  return(FALSE);
}


/*
 * SET SFR addr data...
 */
//FIXME
bool
cmd_set_sfr(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  set_memory(uc->SFR, SFR_SIZE, SFR_START, uc, sfr_tab, sim);
  return(FALSE);
}


/*
 * SET BIT addr data...
 */
//FIXME
bool
cmd_set_bit(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  uchar *cell, addr;
  uchar bitaddr, bitmask;

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Address has not given.\n");
      return(FALSE);
    }
  if (!interpret_bitname(s, uc, &cell, &addr, &bitaddr, &bitmask, NULL))
    {
      fprintf(/*uc*/sim->cmd_out, "Bad address %s\n", s);
      return(FALSE);
    }

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Data has not given.\n");
      return(FALSE);
    }
  while (s)
    {
      if (!isdigit(*s))
	{
	  fprintf(/*uc*/sim->cmd_out, "Bad data %s\n", s);
	  return(FALSE);
	}
      if (*s == '0')
	*cell&= ~bitmask;
      else
	*cell|= bitmask;
      bitmask<<= 1;
      bitmask&= 0xff;
      s= strtok(NULL, delimiters);
    }
  return(FALSE);
}


/*
 * SET PORT port data
 */
//FIXME
bool
cmd_set_port(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s, *p;
  uchar port, data;
  
  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Port number has not given.\n");
      return(FALSE);
    }
  port= strtol(s, &p, 0);
  if ((p && *p) ||
      (port > 3))
    {
      fprintf(/*uc*/sim->cmd_out, "Port number %s is wrong.\n", s);
      return(FALSE);
    }

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      fprintf(/*uc*/sim->cmd_out, "Date has not given.\n");
      return(FALSE);
    }
  data= strtol(s, &p, 0);
  if (p && *p)
    {
      fprintf(/*uc*/sim->cmd_out, "Data %s is wrong.\n", s);
      return(FALSE);
    }
  uc->port_pins[port]= data;
  return(FALSE);
}


/*
 * Fill memory commands
 */
//FIXME
static void
fill_memory(uchar *what, int size, int first, class t_uc51 *uc,
	    class cl_sim *sim)
{
  char *s, *p;
  int start, stop;
  uchar data;

  if ((s= strtok(NULL, delimiters)) == NULL)
    return;
  start= strtol(s, &p, 0);
  if ((p && *p) ||
      (start < 0) ||
      (start < first) ||
      (start >= size))
    fprintf(sim->cmd_out, "Start address %s is wrong.\n", s);

  if ((s= strtok(NULL, delimiters)) == NULL)
    return;
  stop= strtol(s, &p, 0);
  if ((p && *p) ||
      (stop < start) ||
      (stop >= size))
    fprintf(sim->cmd_out, "Stop address %s is wrong.\n", s);

  if ((s= strtok(NULL, delimiters)) == NULL)
    return;
  data= strtol(s, &p, 0);
  if (p && *p)
    fprintf(sim->cmd_out, "Data %s is wrong.\n", s);
  
  while (start <= stop)
    {
      what[start]= data;
      if (what == uc->SFR)
	{
	  uc->proc_write(&what[start]);
	  if (start == ACC)
	    uc->set_p_flag();
	}
      if (what == uc->MEM(MEM_ROM))
	uc->del_inst_at(start);
      start++;
    }
}


/*
 * FILL IRAM from to data
 */
//FIXME
bool
cmd_fill_iram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  fill_memory(uc->IRAM, IRAM_SIZE, 0, uc, sim);
  return(FALSE);
}


/*
 * FILL XRAM from to data
 */
//FIXME
bool
cmd_fill_xram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  fill_memory(uc->MEM(MEM_XRAM), uc->get_mem_size(MEM_XRAM), 0, uc, sim);
  uc->xram2eram();
  return(FALSE);
}


/*
 * FILL SFR from to data
 */
//FIXME
bool
cmd_fill_sfr(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  fill_memory(uc->SFR, SFR_SIZE, SFR_START, uc, sim);
  return(FALSE);
}


/*
 * FILL CODE from to data
 */
//FIXME
bool
cmd_fill_code(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  fill_memory(uc->MEM(MEM_ROM), EROM_SIZE, 0, uc, sim);
  return(FALSE);
}


/* End of set.cc */
