/*
 * Simulator of MCS51 (arith.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>

#include "uc51cl.h"
#include "stypes.h"


/*
 * 0x03 1 12 RR A
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_rr(uchar code)
{
  if (SFR[event_at.ws= event_at.rs= ACC] & 0x01)
    SFR[ACC]= (SFR[ACC] >> 1) & 0x80;
  else
    SFR[ACC]>>= 1;
  return(resGO);
}


/*
 * 0x13 1 12 RRC A
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_rrc(uchar code)
{
  bool cy;

  cy= GET_C;
  SET_C(SFR[ACC] & 0x01);
  SFR[event_at.ws= event_at.rs= ACC]>>= 1;
  if (cy)
    SFR[ACC]|= 0x80;
  return(resGO);
}


/*
 * 0x23 1 12 RL A
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_rl(uchar code)
{
  if (SFR[event_at.ws= event_at.rs= ACC] & 0x80)
    SFR[ACC]= (SFR[ACC] << 1 ) | 0x01;
  else
    SFR[ACC]<<= 1;
  return(resGO);
}


/*
 * 0x24 2 12 ADD A,#data
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_add_a_$data(uchar code)
{
  uchar data;
  bool newC, newA, c6;

  data= fetch();
  newC= (((uint)SFR[ACC]+(uint)(data)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)) & 0x80;
  SFR[event_at.ws= ACC]+= data;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x25 2 12 ADD A,addr
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_add_a_addr(uchar code)
{
  uchar data;
  bool newC, newA, c6;

  data= read(get_direct(fetch(), &event_at.ri, &event_at.rs));
  newC= (((uint)SFR[ACC]+(uint)(data)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)) & 0x80;
  SFR[event_at.ws= ACC]+= data;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x26-0x27 1 12 ADD A,@Ri
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_add_a_$ri(uchar code)
{
  uchar data, *addr;
  bool newC, newA, c6;
  int res;

  addr= get_indirect(event_at.ri= *(get_reg(code & 0x01)), &res);
  data= *addr;
  newC= (((uint)SFR[ACC]+(uint)data) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)) & 0x80;
  SFR[event_at.ws= ACC]+= data;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(res);
}


/*
 * 0x28-0x2f 1 12 ADD A,Rn
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_add_a_rn(uchar code)
{
  uchar data;
  bool newC, newA, c6;

  data= *(get_reg(code & 0x07, &event_at.ri));
  newC= (((uint)SFR[ACC]+(uint)data) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)) & 0x80;
  SFR[event_at.ws= ACC]+= data;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x33 1 12 RLC A
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_rlc(uchar code)
{
  bool cy;

  cy= GET_C;
  SET_C(SFR[ACC] & 0x80);
  SFR[event_at.ws= event_at.rs= ACC]<<= 1;
  if (cy)
    SFR[ACC]|= 0x01;
  return(resGO);
}


/*
 * 0x34 2 12 ADDC A,#data
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_addc_a_$data(uchar code)
{
  uchar data;
  bool orgC, newC, newA, c6;

  data= fetch();
  newC= (((uint)SFR[ACC]+(uint)data+((orgC= GET_C)?1:0)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)+(orgC?1:0)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)+(orgC?1:0)) & 0x80;
  SFR[event_at.ws= ACC]= SFR[ACC] + data + (orgC?1:0);
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x35 2 12 ADDC A,addr
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_addc_a_addr(uchar code)
{
  uchar data;
  bool orgC, newC, newA, c6;

  data= read(get_direct(fetch(), &event_at.ri, &event_at.rs));
  newC= (((uint)SFR[ACC]+(uint)data+((orgC= GET_C)?1:0)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)+(orgC?1:0)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)+(orgC?1:0)) & 0x80;
  SFR[event_at.ws= ACC]= SFR[ACC] + data + (orgC?1:0);
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x36-0x37 1 12 ADDC A,@Ri
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_addc_a_$ri(uchar code)
{
  uchar data, *addr;
  bool orgC, newC, newA, c6;
  int res;

  addr= get_indirect(event_at.ri= *(get_reg(code & 0x01)), &res);
  data= *addr;
  newC= (((uint)SFR[ACC]+(uint)data+((orgC= GET_C)?1:0)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)+(orgC?1:0)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)+(orgC?1:0)) & 0x80;
  SFR[event_at.ws= ACC]= SFR[ACC] + data + (orgC?1:0);
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(res);
}


/*
 * 0x38-0x3f 1 12 ADDC A,Rn
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_addc_a_rn(uchar code)
{
  uchar data;
  bool orgC, newC, newA, c6;

  data= *(get_reg(code & 0x07, &event_at.ri));
  newC= (((uint)SFR[ACC]+(uint)data+((orgC= GET_C)?1:0)) > 255)?0x80:0;
  newA= ((SFR[ACC]&0x0f)+(data&0x0f)+(orgC?1:0)) & 0xf0;
  c6  = ((SFR[ACC]&0x7f)+(data&0x7f)+(orgC?1:0)) & 0x80;
  SFR[event_at.ws= ACC]= SFR[ACC] + data + (orgC?1:0);
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x84 1 48 DIV AB
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_div_ab(uchar code)
{
  uchar temp;

  SFR[PSW]&= ~bmCY;
  if (!SFR[event_at.rs= B])
    SFR[PSW]|= bmOV;
  else
    {
      SFR[PSW]&= ~bmOV;
      temp= SFR[ACC] / SFR[B];
      SFR[B]= SFR[ACC] % SFR[B];
      SFR[event_at.ws= ACC]= temp;
    }
  tick(3);
  return(resGO);
}


/*
 * 0x94 2 12 SUBB A,#data
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_subb_a_$data(uchar code)
{
  uchar data, d;
  bool newC, newA, c6;

  data= fetch();
  d= ~data + (GET_C?0:1);
  newC= (SFR[ACC] < data+(GET_C?1:0))?0x80:0;
  newA= !(((SFR[ACC]&0x0f)+(d&0x0f)) & 0xf0);
  c6  = (((SFR[ACC]&0x7f)+(d&0x7f)) & 0x80)?0:0x80;
  SFR[event_at.ws= ACC]-= data;
  if (GET_C)
    SFR[ACC]--;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x95 2 12 SUBB A,addr
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_subb_a_addr(uchar code)
{
  uchar *addr, data, d;
  bool newC, newA, c6;

  addr= get_direct(fetch(), &event_at.ri, &event_at.rs);
  data= read(addr);
  d= ~data + (GET_C?0:1);
  newC= (SFR[ACC] < data+(GET_C?1:0))?0x80:0;
  newA= !(((SFR[ACC]&0x0f)+(d&0x0f)) & 0xf0);
  c6  = (((SFR[ACC]&0x7f)+(d&0x7f)) & 0x80)?0:0x80;
  SFR[event_at.ws= ACC]-= data;
  if (GET_C)
    SFR[ACC]--;
  SET_C(newC);
  SET_BIT(newC ^ c6, PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0x96-0x97 1 12 SUBB A,@Ri
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_subb_a_$ri(uchar code)
{
  uchar data, d;
  bool newC, newA, c6;
  int res;

  data= *(get_indirect(event_at.ri= *(get_reg(code & 0x01)), &res));
  d= ~data + (GET_C?0:1);
  newC= (SFR[ACC] < data+(GET_C?1:0))?0x80:0;
  newA= !(((SFR[ACC]&0x0f)+(d&0x0f)) & 0xf0);
  c6  = (((SFR[ACC]&0x7f)+(d&0x7f)) & 0x80)?0:0x80;
  SFR[event_at.ws= ACC]-= data;
  if (GET_C)
    SFR[ACC]--;
  SET_C(newC);
  SET_BIT(newC ^ (SFR[ACC] & 0x80), PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(res);
}


/*
 * 0x98-0x9f 1 12 SUBB A,Rn
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_subb_a_rn(uchar code)
{
  uchar data, d;
  bool newC, newA, c6;

  data= *(get_reg(code & 0x07, &event_at.ri));
  d= ~data + (GET_C?0:1);
  newC= (SFR[ACC] < data+(GET_C?1:0))?0x80:0;
  newA= !(((SFR[ACC]&0x0f)+(d&0x0f)) & 0xf0);
  c6  = (((SFR[ACC]&0x7f)+(d&0x7f)) & 0x80)?0:0x80;
  SFR[event_at.ws= ACC]-= data;
  if (GET_C)
    SFR[ACC]--;
  SET_C(newC);
  SET_BIT(newC ^ (SFR[ACC] & 0x80), PSW, bmOV);
  SET_BIT(newA, PSW, bmAC);
  return(resGO);
}


/*
 * 0xa4 1 48 MUL AB
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_mul_ab(uchar code)
{
  uint temp;

  SFR[PSW]&= ~bmCY;
  temp= SFR[ACC] * SFR[B];
  SFR[event_at.ws= ACC]= temp & 0xff;
  SFR[event_at.rs= B]  = (temp >> 8) & 0xff;
  SET_BIT(SFR[B], PSW, bmOV);
  tick(3);
  return(resGO);
}


/*
 * 0xd4 1 12 DA A
 *____________________________________________________________________________
 *
 */

int
t_uc51::inst_da_a(uchar code)
{
  if ((SFR[ACC] & 0x0f) > 9 ||
      (SFR[PSW] & bmAC))
    {
      if (((uint)SFR[ACC]+(uint)0x06) > 255)
	SFR[PSW]|= bmCY;
      SFR[event_at.ws= ACC]+= 0x06;
    }
  if ((SFR[ACC] & 0xf0) > 0x90 ||
      (SFR[PSW] & bmCY))
    {
      if (((uint)SFR[ACC]+(uint)0x60) > 255)
	SFR[PSW]|= bmCY;
      SFR[event_at.ws= ACC]+= 0x60;
    }
  return(resGO);
}


/* End of arith.cc */
